#include "OpenGLDrv.h"

enum EDrawnObject
{
	DO_None,
	DO_Tile,
	DO_MeshPoly,
	DO_BspPoly
};

enum EStyle
{
	Normal,
	Masked,
	Translucent,
	Modulated,
	Multiply,
	Alpha,
	Warp
};

// Can't store just FTextureInfo pointers because they quickly get deleted.
// Also can't store lightmap pointers because the lightmap often gets overwritten.

struct LightMapInfo
{
	bool HasLightMap;
	int BufferPos;
	int USize, VSize;
};

const int LightMapBufferSize = 4*1024*1024;
const int MaxDrawnObjects    = 4096;

int LightMapBuffer[LightMapBufferSize];
int LightMapBufferPos = 0;

struct Tri { float X1, Y1, X2, Y2, X3, Y3; };

struct TriColor { int C1, C2, C3; };

struct TriDepth { float Z1, Z2, Z3; };

struct DrawnObject
{
	EDrawnObject Kind;
	FTextureInfo Texture;
	int PolyFlags;
	int Index;
	
	union
	{
		struct
		{
			float X, Y, XL, YL, U, V, UL, VL;
			float Z;
			int Color;
		} Tile;
		
		struct
		{
			Tri Coords;
			Tri UVs;
			TriColor Colors;
			float Z;
		} MeshPoly;
		
		struct
		{
			LightMapInfo LightMap;
			Tri Coords;
			Tri UVs;
			Tri LightUVs;
			TriDepth Depths;
		} BspPoly;
	};

	float GetDepthHint()
	{
		switch(Kind)
		{
		case DO_Tile:
			return Tile.Z;
			break;
		case DO_MeshPoly:
			return MeshPoly.Z;
			break;
		case DO_BspPoly:
			{
				// calibrate this

				/*float min = BspPoly.Depths.Z1;
				if(BspPoly.Depths.Z2 < min) min = BspPoly.Depths.Z2;
				if(BspPoly.Depths.Z3 < min) min = BspPoly.Depths.Z3;*/

				float max = BspPoly.Depths.Z1;
				if(BspPoly.Depths.Z2 > max) max = BspPoly.Depths.Z2;
				if(BspPoly.Depths.Z3 > max) max = BspPoly.Depths.Z3;

				float avg = (BspPoly.Depths.Z1+BspPoly.Depths.Z2+BspPoly.Depths.Z3)/3;
				
				return max*0.6 + avg*0.4;
			}
			break;
		default:
			return 0;
		}
	}
};

DrawnObject DrawnObjects[MaxDrawnObjects];
DrawnObject* DrawOrder[MaxDrawnObjects];
int NumDrawnObjects;

// is this needed at all?
HWND hWnd;
HDC hDC;

// my external dll stuff
HINSTANCE Dll=0;
typedef void (_stdcall *DTInit       ) (void*,int,int,int); DTInit DInit;
typedef void (_stdcall *DTUnload     ) ();					DTUnload DUnload;
//typedef void (_stdcall *DTPause) (int);					   DTPause DPause;
typedef void (_stdcall *DTBeginUpdate) ();				    DTBeginUpdate DBeginUpdate;
typedef void (_stdcall *DTEndUpdate  ) ();				    DTEndUpdate DEndUpdate;

//typedef void (_stdcall *DTDrawClear) (int);				   DTDrawClear DDrawClear;
//typedef void (_stdcall *DTDrawRect) (int,int,int,int,int); DTDrawRect DDrawRect;
//typedef void (_stdcall *DTDrawLine) (int,int,int,int,int); DTDrawLine DDrawLine;

typedef void (_stdcall *DTSetTexture  ) (BYTE*,int,int,FColor*,int); DTSetTexture DSetTexture;
typedef void (_stdcall *DTSetLightmap ) (void*,int,int            ); DTSetLightmap DSetLightmap;

typedef void (_stdcall *DTDrawTile    ) (float,float,float,float,
									     float,float,float,float,
										 float,float,int          ); DTDrawTile DDrawTile;
typedef void (_stdcall *DTDrawTri     ) (Tri*,Tri*,TriColor*	  ); DTDrawTri DDrawTri;
typedef void (_stdcall *DTDrawPerspTri) (Tri*,Tri*,Tri*,TriDepth* ); DTDrawPerspTri DDrawPerspTri;

typedef void (_stdcall *DTFlashAdd    ) (int);	DTFlashAdd DFlashAdd;
typedef void (_stdcall *DTFlashSub    ) (int);	DTFlashSub DFlashSub;
typedef void (_stdcall *DTFlashMod    ) (int);	DTFlashMod DFlashMod;

#define LOADPROC(name,rawname) \
	D##name = (DT##name) GetProcAddress(Dll,rawname);\
	if(!D##name) return 0;

class UOpenGLRenderDevice : public URenderDevice
{
	DECLARE_CLASS(UOpenGLRenderDevice,URenderDevice,CLASS_Config)
	
	FPlane FlashScale, FlashFog;
	
	void AddDrawnObject(DrawnObject& obj)
	{
		if(NumDrawnObjects<MaxDrawnObjects)
		{
			appMemcpy(DrawnObjects+NumDrawnObjects, &obj, sizeof(DrawnObject));
			DrawnObjects[NumDrawnObjects].Index = NumDrawnObjects;
			DrawOrder[NumDrawnObjects] = &DrawnObjects[NumDrawnObjects];
			NumDrawnObjects++;
		}
	}

	static int DrawnObjectCompare(DrawnObject** obj1, DrawnObject** obj2)
	{
		if((*obj1)->Kind==DO_Tile && (*obj2)->Kind==DO_Tile)
			return (*obj1)->Index - (*obj2)->Index;

		if((*obj1)->GetDepthHint() > (*obj2)->GetDepthHint())
			return -1;
		else
			return 1;
	}

	void SortDrawnObjects()
	{
		appQsort(DrawOrder, NumDrawnObjects, sizeof(DrawnObject*), (QSORT_COMPARE)DrawnObjectCompare);
	}

	int BufferLightMap(FTextureInfo* litmap)
	{
		int pixels = litmap->USize*litmap->VSize;

		if(LightMapBufferPos+pixels <= LightMapBufferSize)
		{
			int startPos = LightMapBufferPos;
			appMemcpy(LightMapBuffer+LightMapBufferPos, litmap->Mips[0]->DataPtr, pixels*4);
			LightMapBufferPos += pixels;
			return startPos;
		}
		else
			return -1;
	}
	
	// URenderDevice interface.
	UBOOL Init( UViewport* InViewport, INT NewX, INT NewY, INT NewColorBytes, UBOOL Fullscreen )
	{
		guard(UOpenGLRenderDevice::Init);
		
		// Driver flags.
		SpanBased			= 0;
		FullscreenOnly		= 0;
		SupportsFogMaps		= 1;
		SupportsDistanceFog	= 1;
		
		//load dll
		Dll = LoadLibraryA("YRaster.dll");
		
		if(!Dll)
			return 0;
		
		LOADPROC(Init        , "INIT@16"        )
		LOADPROC(Unload      , "UNLOAD@0"       )
		LOADPROC(BeginUpdate , "BEGINUPDATE@0"  )
		LOADPROC(EndUpdate   , "ENDUPDATE@0"    )
		LOADPROC(SetTexture  , "SETTEXTURE@20"  )
		LOADPROC(SetLightmap , "SETLIGHTMAP@12" )
		LOADPROC(DrawTile    , "DRAWTILE@44"    )
		LOADPROC(DrawTri     , "DRAWTRI@12"     )
		LOADPROC(DrawPerspTri, "DRAWPERSPTRI@16")
		LOADPROC(FlashAdd    , "FLASHADD@4"     )
		LOADPROC(FlashSub    , "FLASHSUB@4"     )
		LOADPROC(FlashMod    , "FLASHMOD@4"     )

		// is this neccessary?
		Viewport = InViewport;
		hWnd = (HWND)Viewport->GetWindow();
		hDC = GetDC(hWnd);

		NumDrawnObjects = 0;
		LightMapBufferPos = 0;
		
		SetRes(NewX,NewY,NewColorBytes,Fullscreen);
		
		//Viewport->Stride == 0 ???
		//DInit(Viewport->ScreenPointer,NewX,NewY,NewX*4);

		return 1;
		unguard;
	}
	void Exit()
	{
		guard(UOpenGLRenderDevice::Exit);

		// is this necessary?
		if( hDC )
			ReleaseDC(hWnd,hDC);

		if(Dll)
		{
			DUnload();
			FreeLibrary(Dll);
		}
		unguard;
	}
	void UnsetRes()
	{
		guard(UOpenGLRenderDevice::UnsetRes);
		unguard;
	}
	void PrintFormat( HDC hDC, INT nPixelFormat )
	{
		guard(UOpenGlRenderDevice::PrintFormat);
		unguard;
	}
	UBOOL SetRes( INT NewX, INT NewY, INT NewColorBytes, UBOOL Fullscreen )
	{
		guard(UOpenGlRenderDevice::SetRes);
		
		//Viewport->ResizeViewport( Fullscreen ? (BLIT_Fullscreen) : (BLIT_HardwarePaint), NewX, NewY, NewColorBytes );
		Viewport->ResizeViewport( Fullscreen ? (BLIT_Fullscreen | BLIT_DibSection) : (BLIT_HardwarePaint | BLIT_DibSection), NewX, NewY, NewColorBytes );
		DInit(Viewport->ScreenPointer,NewX,NewY,NewX*4);

		if(Fullscreen)
		{
			DEVMODEA dm;
			ZeroMemory( &dm, sizeof(dm) );
			dm.dmSize       = sizeof(dm);
			dm.dmPelsWidth  = NewX;
			dm.dmPelsHeight = NewY;
			dm.dmBitsPerPel = NewColorBytes;
			dm.dmFields     = DM_PELSWIDTH | DM_PELSHEIGHT;
			//DM_BITSPERPEL
			if( ChangeDisplaySettingsA( &dm, CDS_FULLSCREEN )!=DISP_CHANGE_SUCCESSFUL )
			{
				debugf( TEXT("ChangeDisplaySettingsA failed: %ix%i"), NewX, NewY );
				return 0;
			}
		}

		return 0;
		unguard;
	}
	void ShutdownAfterError()
	{
		guard(UOpenGLRenderDevice::ShutdownAfterError);
		debugf( NAME_Exit, TEXT("UOpenGLRenderDevice::ShutdownAfterError") );
		ChangeDisplaySettings( NULL, 0 );
		DUnload();
		unguard;
	}
	void Flush()
	{
		guard(UOpenGLRenderDevice::Flush);
		unguard;
	}
	UBOOL Exec( const TCHAR* Cmd, FOutputDevice& Ar )
	{
		guard(UOpenGLRenderDevice::Exec);	
		return 0;
		unguard;
	}
	void Lock( FPlane InFlashScale, FPlane InFlashFog, FPlane ScreenClear, DWORD RenderLockFlags, BYTE* InHitData, INT* InHitSize )
	{
		guard(UOpenGLRenderDevice::Lock);
		//DDrawClear(0);
		NumDrawnObjects = 0;
		LightMapBufferPos = 0;
		FlashScale = InFlashScale;
		FlashFog = InFlashFog;
		unguard;
	}
	void SetSceneNode( FSceneNode* Frame )
	{
		guard(UOpenGLDriver::SetSceneNode);
		unguard;
	}
	bool IsPowerOfTwo(int x)
	{
		for(int i=1;i<=1024;i = i << 1)
		{
			if(x==i)
				return true;
		}
		return false;
	}
	void ValidateTex(FTextureInfo* tex)
	{
		TCHAR msg[64];

		if(tex)
		{
			if(IsPowerOfTwo(tex->USize) && IsPowerOfTwo(tex->VSize))
				return;
			else
				appSprintf(msg,L"Bad texture size: %ix%i",tex->USize,tex->VSize);
		}
		else
			return;
		
		MessageBox(NULL,msg,L"Error",0);
	}
	void Unlock( UBOOL Blit )
	{
		guard(UOpenGLRenderDevice::Unlock);

		if(!Blit)
			return;

		SortDrawnObjects();
		
		DBeginUpdate();
		for(int i=0;i<NumDrawnObjects;i++)
		{
			DrawnObject& obj = *DrawOrder[i];

			switch(obj.Kind)
			{
			case DO_Tile:
				ValidateTex(&obj.Texture);

				DSetTexture(obj.Texture.Mips[0]->DataPtr,
							obj.Texture.USize,
							obj.Texture.VSize, 
							obj.Texture.Palette, 
							Style(obj.PolyFlags));
				DDrawTile(obj.Tile.X,
						  obj.Tile.Y,
						  obj.Tile.XL,
						  obj.Tile.YL,
						  obj.Tile.U,
						  obj.Tile.V,
						  obj.Tile.UL,
						  obj.Tile.VL,
						  obj.Texture.UClamp,
						  obj.Texture.VClamp,
						  obj.Tile.Color);
				break;
			case DO_MeshPoly:
				DSetTexture(obj.Texture.Mips[0]->DataPtr,
							obj.Texture.USize,
							obj.Texture.VSize, 
							obj.Texture.Palette, 
							Style(obj.PolyFlags));
				DDrawTri(&obj.MeshPoly.Coords,
						 &obj.MeshPoly.UVs,
						 &obj.MeshPoly.Colors);
				break;
			case DO_BspPoly:
				DSetTexture(obj.Texture.Mips[0]->DataPtr,
							obj.Texture.USize,
							obj.Texture.VSize, 
							obj.Texture.Palette, 
							Style(obj.PolyFlags));

				if(!obj.BspPoly.LightMap.HasLightMap)
					DSetLightmap(0,0,0);
				else
					DSetLightmap(LightMapBuffer+obj.BspPoly.LightMap.BufferPos,
								 obj.BspPoly.LightMap.USize,
								 obj.BspPoly.LightMap.VSize);

				DDrawPerspTri(&obj.BspPoly.Coords,
							  &obj.BspPoly.UVs,
							  &obj.BspPoly.LightUVs,
							  &obj.BspPoly.Depths);
				break;
			}
		}
		//if( FlashScale!=FPlane(.5,.5,.5,0) || FlashFog!=FPlane(0,0,0,0) )
		//{
		//	DFlashAdd(ConvertColor(FlashFog));
		//}
		DEndUpdate();

		unguard;
	}
	void DrawComplexSurface( FSceneNode* Frame, FSurfaceInfo& Surface, FSurfaceFacet& Facet )
	{
		guard(UOpenGLRenderDevice::DrawComplexSurface);

		FLOAT UDot = Facet.MapCoords.XAxis | Facet.MapCoords.Origin;
		FLOAT VDot = Facet.MapCoords.YAxis | Facet.MapCoords.Origin;

		for(FSavedPoly* Poly = Facet.Polys; Poly != NULL; Poly = Poly->Next)
		{
			static float Us[32];
			static float Vs[32];
			static float LightUs[32];
			static float LightVs[32];
			
			//calc UVs
			for(int i=0; i<Poly->NumPts; i++)
			{
				
				FLOAT U = Facet.MapCoords.XAxis | Poly->Pts[i]->Point;
				FLOAT V = Facet.MapCoords.YAxis | Poly->Pts[i]->Point;

				Us[i]      = (U-UDot-Surface.Texture->Pan.X)/Surface.Texture->UScale;
				Vs[i]      = (V-VDot-Surface.Texture->Pan.Y)/Surface.Texture->VScale;
				if(Surface.LightMap)
				{
					LightUs[i] = (U-UDot-Surface.LightMap->Pan.X)/Surface.LightMap->UScale;
					LightVs[i] = (V-VDot-Surface.LightMap->Pan.Y)/Surface.LightMap->VScale;
				}
			}
			
			for(int j=1; j<Poly->NumPts-1; j++)
			{
				FTransform* Pt1 = Poly->Pts[0];
				FTransform* Pt2 = Poly->Pts[j];
				FTransform* Pt3 = Poly->Pts[j+1];
				
				DrawnObject obj;

				obj.Kind = DO_BspPoly;
				obj.PolyFlags = Surface.PolyFlags;
				appMemcpy(&obj.Texture,Surface.Texture,sizeof(FTextureInfo));
				
				if(Surface.LightMap)
				{
					int buffPos = BufferLightMap(Surface.LightMap);
					
					if(buffPos != -1)
					{
						obj.BspPoly.LightMap.HasLightMap = true;
						obj.BspPoly.LightMap.BufferPos = buffPos;
						obj.BspPoly.LightMap.USize = Surface.LightMap->USize;
						obj.BspPoly.LightMap.VSize = Surface.LightMap->VSize;
					}
					else
					{
						// buffer is full
						obj.BspPoly.LightMap.HasLightMap = false;
					}
				}
				else
				{
					obj.BspPoly.LightMap.HasLightMap = false;
				}
				obj.BspPoly.Coords.X1 = Pt3->ScreenX;
				obj.BspPoly.Coords.Y1 = Pt3->ScreenY;
				obj.BspPoly.Depths.Z1 = Pt3->Point.Z;
				obj.BspPoly.Coords.X2 = Pt2->ScreenX;
				obj.BspPoly.Coords.Y2 = Pt2->ScreenY;
				obj.BspPoly.Depths.Z2 = Pt2->Point.Z;
				obj.BspPoly.Coords.X3 = Pt1->ScreenX;
				obj.BspPoly.Coords.Y3 = Pt1->ScreenY;
				obj.BspPoly.Depths.Z3 = Pt1->Point.Z;
				obj.BspPoly.UVs.X1 = Us[j+1];
				obj.BspPoly.UVs.Y1 = Vs[j+1];
				obj.BspPoly.UVs.X2 = Us[j];
				obj.BspPoly.UVs.Y2 = Vs[j];
				obj.BspPoly.UVs.X3 = Us[0];
				obj.BspPoly.UVs.Y3 = Vs[0];
				obj.BspPoly.LightUVs.X1 = LightUs[j+1];
				obj.BspPoly.LightUVs.Y1 = LightVs[j+1];
				obj.BspPoly.LightUVs.X2 = LightUs[j];
				obj.BspPoly.LightUVs.Y2 = LightVs[j];
				obj.BspPoly.LightUVs.X3 = LightUs[0];
				obj.BspPoly.LightUVs.Y3 = LightVs[0];

				AddDrawnObject(obj);
			}
		}
		unguard;
	}
	EStyle Style(DWORD PolyFlags)
	{
		if(PolyFlags & PF_Modulated)   return Modulated;
		if(PolyFlags & PF_Translucent) return Translucent;
		if(PolyFlags & PF_Masked)      return Masked;
		return Normal;
	}
	int ConvertColor(FPlane col)
	{
		if(col.X>1) col.X=1;
		if(col.X<0) col.X=0;
		if(col.Y>1) col.Y=1;
		if(col.Y<0) col.Y=0;
		if(col.Z>1) col.Z=1;
		if(col.Z<0) col.Z=0;

		return ((int(col.X*255)&255)<<0) | ((int(col.Y*255)&255)<<8) | ((int(col.Z*255)&255)<<16);
	}
	int ConvertColorSwap(FPlane col)
	{
		if(col.X>1) col.X=1;
		if(col.X<0) col.X=0;
		if(col.Y>1) col.Y=1;
		if(col.Y<0) col.Y=0;
		if(col.Z>1) col.Z=1;
		if(col.Z<0) col.Z=0;

		return ((int(col.X*255)&255)<<16) | ((int(col.Y*255)&255)<<8) | ((int(col.Z*255)&255)<<0);
	}
	void DrawGouraudPolygon( FSceneNode* Frame, FTextureInfo& Info, FTransTexture** Pts, INT NumPts, DWORD PolyFlags, FSpanBuffer* Span )
	{
		guard(UOpenGLRenderDevice::DrawGouraudPolygon);

		for(int i=1;i<NumPts-1;i++)
		{
			FTransTexture* Pt1 = Pts[i+1];
			FTransTexture* Pt2 = Pts[i];
			FTransTexture* Pt3 = Pts[0];

			DrawnObject obj;
			
			obj.Kind        = DO_MeshPoly;
			appMemcpy(&obj.Texture,&Info,sizeof(FTextureInfo));
			obj.PolyFlags   = PolyFlags;
			obj.MeshPoly.Coords.X1 = Pt1->ScreenX;
			obj.MeshPoly.Coords.Y1 = Pt1->ScreenY;
			obj.MeshPoly.Coords.X2 = Pt2->ScreenX;
			obj.MeshPoly.Coords.Y2 = Pt2->ScreenY;
			obj.MeshPoly.Coords.X3 = Pt3->ScreenX;
			obj.MeshPoly.Coords.Y3 = Pt3->ScreenY;
			obj.MeshPoly.UVs.X1 = Pt1->U;
			obj.MeshPoly.UVs.Y1 = Pt1->V;
			obj.MeshPoly.UVs.X2 = Pt2->U;
			obj.MeshPoly.UVs.Y2 = Pt2->V;
			obj.MeshPoly.UVs.X3 = Pt3->U;
			obj.MeshPoly.UVs.Y3 = Pt3->V;
			obj.MeshPoly.Colors.C1 = ConvertColor(Pt1->Light);
			obj.MeshPoly.Colors.C2 = ConvertColor(Pt2->Light);
			obj.MeshPoly.Colors.C3 = ConvertColor(Pt3->Light);
			obj.MeshPoly.Z  = (Pt1->Point.Z+Pt2->Point.Z+Pt3->Point.Z)/3;
			
			AddDrawnObject(obj);
		}

		unguard;
	}
	void DrawTile( FSceneNode* Frame, FTextureInfo& Info, FLOAT X, FLOAT Y, FLOAT XL, FLOAT YL, FLOAT U, FLOAT V, FLOAT UL, FLOAT VL, class FSpanBuffer* Span, FLOAT Z, FPlane Color, FPlane Fog, DWORD PolyFlags )
	{
		guard(UOpenGLRenderDevice::DrawTile);

		DrawnObject obj;
		
		obj.Kind       = DO_Tile;
		appMemcpy(&obj.Texture,&Info,sizeof(FTextureInfo));
		obj.PolyFlags  = PolyFlags;
		obj.Tile.X     = X;
		obj.Tile.Y     = Y;
		obj.Tile.XL    = XL;
		obj.Tile.YL    = YL;
		obj.Tile.U     = U;
		obj.Tile.V     = V;
		obj.Tile.UL    = UL;
		obj.Tile.VL    = VL;
		obj.Tile.Z     = Z;
		obj.Tile.Color = ConvertColorSwap(Color);
		
		AddDrawnObject(obj);

		unguard;
	}
	void Draw2DLine( FSceneNode* Frame, FPlane Color, DWORD LineFlags, FVector P1, FVector P2 )
	{
		guard(UOpenGLRenderDevice::Draw2DLine);
		//DDrawLine(P1.X,P1.Y,P2.X,P2.Y,0x00FF00);
		unguard;
	}
	void Draw3DLine( FSceneNode* Frame, FPlane Color, DWORD LineFlags, FVector P1, FVector P2 )
	{
		guard(UOpenGLRenderDevice::Draw3DLine);
		//DDrawLine(P1.X,P1.Y,P2.X,P2.Y,0xFF0000);
		unguard;
	}
	void Draw2DPoint( FSceneNode* Frame, FPlane Color, DWORD LineFlags, FLOAT X1, FLOAT Y1, FLOAT X2, FLOAT Y2, FLOAT Z )
	{
		guard(UOpenGLRenderDevice::Draw2DPoint);
		unguard;
	}
	void ClearZ( FSceneNode* Frame )
	{
		guard(UOpenGLRenderDevice::ClearZ);
		unguard;
	}
	void PushHit( const BYTE* Data, INT Count )
	{
		guard(UOpenGLRenderDevice::PushHit);
		unguard;
	}
	void PopHit( INT Count, UBOOL bForce )
	{
		guard(UOpenGLRenderDevice::PopHit);
		unguard;
	}
	void GetStats( TCHAR* Result )
	{
		guard(UOpenGLRenderDevice::GetStats);
		/*appSprintf
		(
			Result,
			TEXT("OpenGL stats: Bind=1 Image=1 Complex=1 Gouraud=1 Tile=1")
		);*/
		unguard;
	}
	void ReadPixels( FColor* Pixels )
	{
		guard(UOpenGLRenderDevice::ReadPixels);
		appMemcpy(Pixels,Viewport->ScreenPointer,Viewport->SizeX*4*Viewport->SizeY);
  		unguard;
	}
	void EndFlash()
	{
	}
	void PrecacheTexture( FTextureInfo& Info, DWORD PolyFlags )
	{
		guard(UOpenGLRenderDevice::PrecacheTexture);
		unguard;
	}
};
IMPLEMENT_CLASS(UOpenGLRenderDevice);